# ***************************************************************************
# *   Copyright (c) 2022 Zheng Lei (realthunder) <realthunder.dev@gmail.com>*
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Auto code generator for parameters in Preferences/TreeView"""
import sys
from os import path

# import Tools/params_utils.py
sys.path.append(path.join(path.dirname(path.dirname(path.abspath(__file__))), "Tools"))
import params_utils

from params_utils import (
    ParamBool,
    ParamInt,
    ParamString,
    ParamUInt,
    ParamFloat,
    ParamSpinBox,
    ParamColor,
    ParamHex,
)

NameSpace = "Gui"
ClassName = "TreeParams"
ParamPath = "User parameter:BaseApp/Preferences/TreeView"
ClassDoc = "Convenient class to obtain tree view related parameters"
SourceFile = "Tree.cpp"

Params = [
    ParamBool("SyncSelection", True, on_change=True),
    ParamBool("CheckBoxesSelection", False, on_change=True, title="Show item checkbox"),
    ParamBool("SyncView", True),
    ParamBool("PreSelection", True),
    ParamBool("SyncPlacement", False),
    ParamBool("RecordSelection", True),
    ParamInt("DocumentMode", 2, on_change=True),
    ParamInt("StatusTimeout", 100),
    ParamInt("SelectionTimeout", 100),
    ParamInt("PreSelectionTimeout", 500),
    ParamInt("PreSelectionDelay", 700),
    ParamInt("PreSelectionMinDelay", 200),
    ParamBool("RecomputeOnDrop", True),
    ParamBool("KeepRootOrder", True),
    ParamBool("TreeActiveAutoExpand", True),
    ParamUInt("TreeActiveColor", 0xE6E6FFFF, on_change=True),
    ParamUInt("TreeEditColor", 0x929200FF, on_change=True),
    ParamUInt("SelectingGroupColor", 0x408081FF, on_change=True),
    ParamBool("TreeActiveBold", True, on_change=True),
    ParamBool("TreeActiveItalic", False, on_change=True),
    ParamBool("TreeActiveUnderlined", False, on_change=True),
    ParamBool("TreeActiveOverlined", False, on_change=True),
    ParamInt("Indentation", 0, on_change=True),
    ParamBool("LabelExpression", False),
    ParamInt("IconSize", 0, on_change=True),
    ParamInt(
        "FontSize",
        0,
        on_change=True,
        title="Font size",
        proxy=ParamSpinBox(0, 100, 1, suffix=" pt"),
    ),
    ParamInt("ItemSpacing", 0, on_change=True),
    ParamHex(
        "ItemBackground",
        0,
        on_change=True,
        title="Item background color",
        proxy=ParamColor(),
        doc="Tree view item background. Only effective in overlay.",
    ),
    ParamInt(
        "ItemBackgroundPadding",
        0,
        on_change=True,
        title="Item background padding",
        proxy=ParamSpinBox(0, 100, 1, suffix=" px"),
        doc="Tree view item background padding.",
    ),
    ParamBool(
        "HideColumn",
        True,
        on_change=True,
        title="Hide Description",
        doc="Hide extra tree view column for item description.",
    ),
    ParamBool(
        "HideInternalNames",
        True,
        on_change=True,
        title="Hide Internal Names",
        doc="Hide extra tree view column - Internal Names.",
    ),
    ParamBool(
        "HideScrollBar",
        True,
        title="Hide scroll bar",
        doc="Hide tree view scroll bar in dock overlay.",
    ),
    ParamBool(
        "HideHeaderView",
        True,
        title="Hide header",
        doc="Hide tree view header view in dock overlay.",
    ),
    ParamBool(
        "ResizableColumn",
        False,
        on_change=True,
        title="Resizable columns",
        doc="Allow tree view columns to be manually resized.",
    ),
    ParamInt("ColumnSize1", 0),
    ParamInt("ColumnSize2", 0),
    ParamInt("ColumnSize3", 0),
    ParamBool("TreeToolTipIcon", False, title="Show icon in tool tip"),
    ParamBool(
        "VisibilityIcon",
        True,
        on_change=True,
        title="Show visibility icon",
        doc="Displays an eye icon in front of the tree view items, showing the items visibility status. When clicked the visibility is toggled",
    ),
]


def declare_begin():
    params_utils.declare_begin(sys.modules[__name__])


def declare_end():
    params_utils.declare_end(sys.modules[__name__])


def define():
    params_utils.define(sys.modules[__name__])


params_utils.init_params(Params, NameSpace, ClassName, ParamPath)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2020 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""FreeCAD IFC importer - Multicore version"""

import os
import sys
import time

import FreeCAD
import Arch
import ArchIFC
import Draft

from FreeCAD import Base

from importers import importIFCHelper

# global dicts to store ifc object/freecad object relationships

layers = {}  # ifcid : Draft_Layer
materials = {}  # ifcid : Arch_Material
objects = {}  # ifcid : Arch_Component
subs = {}  # host_ifcid: [child_ifcid,...]
adds = {}  # host_ifcid: [child_ifcid,...]
colors = {}  # objname : (r,g,b)


def open(filename):
    "opens an IFC file in a new document"

    return insert(filename)


def insert(filename, docname=None, preferences=None):
    """imports the contents of an IFC file in the given document"""

    import ifcopenshell
    from ifcopenshell import geom

    # reset global values
    global layers
    global materials
    global objects
    global adds
    global subs
    layers = {}
    materials = {}
    objects = {}
    adds = {}
    subs = {}

    # statistics
    starttime = time.time()  # in seconds
    filesize = os.path.getsize(filename) * 0.000001  # in megabytes
    print("Opening", filename + ",", round(filesize, 2), "Mb")

    # setup ifcopenshell
    if not preferences:
        preferences = importIFCHelper.getPreferences()
    settings = ifcopenshell.geom.settings()
    settings.set(settings.USE_BREP_DATA, True)
    settings.set(settings.SEW_SHELLS, True)
    settings.set(settings.USE_WORLD_COORDS, True)
    if preferences["SEPARATE_OPENINGS"]:
        settings.set(settings.DISABLE_OPENING_SUBTRACTIONS, True)
    if preferences["SPLIT_LAYERS"] and hasattr(settings, "APPLY_LAYERSETS"):
        settings.set(settings.APPLY_LAYERSETS, True)

    # setup document
    if not FreeCAD.ActiveDocument:
        if not docname:
            docname = os.path.splitext(os.path.basename(filename))[0]
        doc = FreeCAD.newDocument(docname)
        doc.Label = docname
        FreeCAD.setActiveDocument(doc.Name)

    # open the file
    ifcfile = ifcopenshell.open(filename)
    progressbar = Base.ProgressIndicator()
    productscount = len(ifcfile.by_type("IfcProduct"))
    progressbar.start("Importing " + str(productscount) + " products...", productscount)
    cores = preferences["MULTICORE"]
    iterator = ifcopenshell.geom.iterator(settings, ifcfile, cores)
    iterator.initialize()
    count = 0

    # process objects
    for item in iterator:
        brep = item.geometry.brep_data
        ifcproduct = ifcfile.by_id(item.guid)
        obj = createProduct(ifcproduct, brep)
        progressbar.next(True)
        writeProgress(count, productscount, starttime)
        count += 1

    # process 2D annotations
    annotations = ifcfile.by_type("IfcAnnotation")
    if annotations:
        print("Processing", str(len(annotations)), "annotations...")
        ifcscale = importIFCHelper.getScaling(ifcfile)
        for annotation in annotations:
            importIFCHelper.createAnnotation(
                annotation, FreeCAD.ActiveDocument, ifcscale, preferences
            )

    # post-processing
    processRelationships()
    storeColorDict()

    # finished
    progressbar.stop()
    FreeCAD.ActiveDocument.recompute()
    endtime = round(time.time() - starttime, 1)
    fs = round(filesize, 1)
    ratio = int(endtime / filesize)
    endtime = "%02d:%02d" % (divmod(endtime, 60))
    writeProgress()  # this cleans the line
    print("Finished importing", fs, "Mb in", endtime, "s, or", ratio, "s/Mb")
    return FreeCAD.ActiveDocument


def writeProgress(count=None, total=None, starttime=None):
    """write progress to console"""

    if not FreeCAD.GuiUp:
        if count is None:
            sys.stdout.write("\r")
            return
        r = count / total
        elapsed = round(time.time() - starttime, 1)
        if r:
            rest = elapsed * ((1 - r) / r)
            eta = "%02d:%02d" % (divmod(rest, 60))
        else:
            eta = "--:--"
        hashes = "#" * int(r * 10) + " " * int(10 - r * 10)
        fstring = "\rImporting " + str(total) + " products [{0}] {1}%, ETA: {2}"
        sys.stdout.write(fstring.format(hashes, int(r * 100), eta))


def createProduct(ifcproduct, brep):
    """creates an Arch object from an IFC product"""

    import Part

    shape = Part.Shape()
    shape.importBrepFromString(brep, False)
    shape.scale(1000.0)  # IfcOpenShell outputs in meters
    if ifcproduct.is_a("IfcSpace"):
        obj = Arch.makeSpace()
    else:
        obj = Arch.makeComponent()
    obj.Shape = shape
    objects[ifcproduct.id()] = obj
    setAttributes(obj, ifcproduct)
    setProperties(obj, ifcproduct)
    createLayer(obj, ifcproduct)
    createMaterial(obj, ifcproduct)
    createModelStructure(obj, ifcproduct)
    setRelationships(obj, ifcproduct)
    setColor(obj, ifcproduct)
    return obj


def setAttributes(obj, ifcproduct):
    """sets the IFC attributes of a component"""

    ifctype = ArchIFC.uncamel(ifcproduct.is_a())
    if ifcproduct.Name:
        obj.Label = ifcproduct.Name
    if ifctype in ArchIFC.IfcTypes:
        obj.IfcType = ifctype
    for attr in dir(ifcproduct):
        if attr in obj.PropertiesList:
            value = getattr(ifcproduct, attr)
            if value:
                try:
                    setattr(obj, attr, value)
                except Exception:
                    pass


def setProperties(obj, ifcproduct):
    """sets the IFC properties of a component"""

    props = obj.IfcProperties
    for prel in ifcproduct.IsDefinedBy:
        if prel.is_a("IfcRelDefinesByProperties"):
            pset = prel.RelatingPropertyDefinition
            if pset.is_a("IfcPropertySet"):
                for prop in pset.HasProperties:
                    if hasattr(prop, "NominalValue"):
                        propname = prop.Name + ";;" + pset.Name
                        v = [p.strip("'") for p in str(prop.NominalValue).strip(")").split(")")]
                        propvalue = ";;".join(v)


def setColor(obj, ifcproduct):
    """sets the color of an object"""

    global colors

    color = importIFCHelper.getColorFromProduct(ifcproduct)
    colors[obj.Name] = color
    if FreeCAD.GuiUp and color:
        obj.ViewObject.ShapeColor = color[:3]


def createLayer(obj, ifcproduct):
    """sets the layer of a component"""

    global layers

    if ifcproduct.Representation:
        for rep in ifcproduct.Representation.Representations:
            for layer in rep.LayerAssignments:
                if not layer.id() in layers:
                    layers[layer.id()] = Draft.make_layer(layer.Name)
                layers[layer.id()].Proxy.addObject(layers[layer.id()], obj)


def createMaterial(obj, ifcproduct):
    """sets the material of a component"""

    global materials

    for association in ifcproduct.HasAssociations:
        if association.is_a("IfcRelAssociatesMaterial"):
            material = association.RelatingMaterial
            if material.is_a("IfcMaterialList"):
                material = material.Materials[0]  # take the first one for now...
            if material.is_a("IfcMaterial"):
                if not material.id() in materials:
                    color = importIFCHelper.getColorFromMaterial(material)
                    materials[material.id()] = Arch.makeMaterial(material.Name, color=color)
                obj.Material = materials[material.id()]


def createModelStructure(obj, ifcobj):
    """sets the parent containers of an IFC object"""

    global objects

    for parent in importIFCHelper.getParents(ifcobj):
        if not parent.id() in objects:
            if parent.is_a("IfcProject"):
                parentobj = Arch.makeProject()
            elif parent.is_a("IfcSite"):
                parentobj = Arch.makeSite()
            else:
                parentobj = Arch.makeBuildingPart()
            setAttributes(parentobj, parent)
            setProperties(parentobj, parent)
            createModelStructure(parentobj, parent)
            objects[parent.id()] = parentobj
        if hasattr(objects[parent.id()].Proxy, "addObject"):
            objects[parent.id()].Proxy.addObject(objects[parent.id()], obj)


def setRelationships(obj, ifcobj):
    """sets additions/subtractions"""

    global adds
    global subs

    if hasattr(ifcobj, "HasOpenings") and ifcobj.HasOpenings:
        for rel in ifcobj.HasOpenings:
            subs.setdefault(ifcobj.id(), []).append(rel.RelatedOpeningElement)

    # TODO: assemblies & booleans


def processRelationships():
    """process all stored relationships"""

    for dom in ((subs, "Subtractions"), (adds, "Additions")):
        for key, vals in dom[0].items():
            if key in objects:
                for val in vals:
                    if val in objects:
                        if hasattr(objects[key], dom[1]):
                            g = getattr(objects[key], dom[1])
                            g.append(val)
                            setattr(objects[key], dom[1], g)


def storeColorDict():
    """stores the color dictionary in the document Meta if non-GUI mode"""

    if colors and not FreeCAD.GuiUp:
        import json

        d = FreeCAD.ActiveDocument.Meta
        d["colordict"] = json.dumps(colors)
        FreeCAD.ActiveDocument.Meta = d



from Base.Metadata import export
from GeometrySurface import GeometrySurface
from typing import Any, Final, Tuple

@export(
    Twin="GeomTrimmedSurface",
    TwinPointer="GeomTrimmedSurface",
    PythonName="Part.RectangularTrimmedSurface",
    FatherInclude="Mod/Part/App/GeometrySurfacePy.h",
    Include="Mod/Part/App/Geometry.h",
    Constructor=True,
)
class RectangularTrimmedSurface(GeometrySurface):
    """
    Describes a portion of a surface (a patch) limited by two values of the
    u parameter in the u parametric direction, and two values of the v parameter in the v parametric
    direction. The domain of the trimmed surface must be within the domain of the surface being trimmed.

    The trimmed surface is defined by:
    - the basis surface, and
    - the values (umin, umax) and (vmin, vmax) which limit it in the u and v parametric directions.

    The trimmed surface is built from a copy of the basis surface. Therefore, when the basis surface
    is modified the trimmed surface is not changed. Consequently, the trimmed surface does not
    necessarily have the same orientation as the basis surface.

    Author: Werner Mayer (wmayer@users.sourceforge.net)
    Licence: LGPL
    """

    BasisSurface: Final[Any] = None
    """Represents the basis surface from which the trimmed surface is derived."""

    def setTrim(self, params: Tuple[float, float, float, float]) -> None:
        """
        setTrim(self, params: (u1, u2, v1, v2)) -> None

        Modifies this patch by changing the trim values applied to the original surface
        """
        ...



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Misc Arch util commands"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Add:
    "the Arch Add command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_Add",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Add", "Add Component"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Add", "Adds the selected components to the active object"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and len(FreeCADGui.Selection.getSelection()) > 1

    def Activated(self):
        import Draft
        import Arch

        sel = FreeCADGui.Selection.getSelection()
        if Draft.getType(sel[-1]) == "Space":
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Add space boundary"))
            FreeCADGui.addModule("Arch")
            FreeCADGui.doCommand(
                "Arch.addSpaceBoundaries( FreeCAD.ActiveDocument."
                + sel[-1].Name
                + ", FreeCADGui.Selection.getSelectionEx() )"
            )
        elif Draft.getType(sel[-1]).startswith("Ifc"):
            FreeCADGui.addModule("nativeifc.ifc_tools")
            for s in sel[:-1]:
                FreeCADGui.doCommand(
                    "nativeifc.ifc_tools.aggregate(FreeCAD.ActiveDocument."
                    + s.Name
                    + ",FreeCAD.ActiveDocument."
                    + sel[-1].Name
                    + ")"
                )
        else:
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Grouping"))
            if not Arch.mergeCells(sel):
                host = sel.pop()
                ss = "["
                for o in sel:
                    if len(ss) > 1:
                        ss += ","
                    ss += "FreeCAD.ActiveDocument." + o.Name
                ss += "]"
                FreeCADGui.addModule("Arch")
                FreeCADGui.doCommand(
                    "Arch.addComponents(" + ss + ",FreeCAD.ActiveDocument." + host.Name + ")"
                )
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


class Arch_Remove:
    "the Arch Add command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_Remove",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Remove", "Remove Component"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Remove",
                "Removes the selected components from their parents, or creates a hole in a component",
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and bool(FreeCADGui.Selection.getSelection())

    def Activated(self):
        import Draft

        sel = FreeCADGui.Selection.getSelection()
        if Draft.getType(sel[-1]) == "Space":
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Remove space boundary"))
            FreeCADGui.addModule("Arch")
            FreeCADGui.doCommand(
                "Arch.removeSpaceBoundaries( FreeCAD.ActiveDocument."
                + sel[-1].Name
                + ", FreeCADGui.Selection.getSelection() )"
            )
        elif Draft.getType(sel[-1]).startswith("Ifc"):
            FreeCADGui.addModule("nativeifc.ifc_tools")
            for s in sel[:-1]:
                FreeCADGui.doCommand(
                    "nativeifc.ifc_tools.aggregate(FreeCAD.ActiveDocument."
                    + s.Name
                    + ",FreeCAD.ActiveDocument."
                    + sel[-1].Name
                    + ",mode='opening')"
                )
        else:
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Ungrouping"))
            if len(sel) > 1:
                host = sel.pop()
                ss = "["
                for o in sel:
                    if len(ss) > 1:
                        ss += ","
                    ss += "FreeCAD.ActiveDocument." + o.Name
                ss += "]"
                FreeCADGui.addModule("Arch")
                FreeCADGui.doCommand(
                    "Arch.removeComponents(" + ss + ",FreeCAD.ActiveDocument." + host.Name + ")"
                )
            else:
                FreeCADGui.addModule("Arch")
                FreeCADGui.doCommand(
                    "Arch.removeComponents(FreeCAD.ActiveDocument." + sel[0].Name + ")"
                )
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


class Arch_SplitMesh:
    "the Arch SplitMesh command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_SplitMesh",
            "MenuText": QT_TRANSLATE_NOOP("Arch_SplitMesh", "Split Mesh"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_SplitMesh", "Splits selected meshes into independent components"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and bool(FreeCADGui.Selection.getSelection())

    def Activated(self):
        import Arch

        if FreeCADGui.Selection.getSelection():
            sel = FreeCADGui.Selection.getSelection()
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Split Mesh"))
            for obj in sel:
                n = obj.Name
                nobjs = Arch.splitMesh(obj)
                if len(nobjs) > 1:
                    g = FreeCAD.ActiveDocument.addObject("App::DocumentObjectGroup", n)
                    for o in nobjs:
                        g.addObject(o)
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()


class Arch_MeshToShape:
    "the Arch MeshToShape command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_MeshToShape",
            "MenuText": QT_TRANSLATE_NOOP("Arch_MeshToShape", "Mesh to Shape"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_MeshToShape", "Turns selected meshes into Part shape objects"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and bool(FreeCADGui.Selection.getSelection())

    def Activated(self):
        import Arch
        from draftutils import params

        if FreeCADGui.Selection.getSelection():
            f = FreeCADGui.Selection.getSelection()[0]
            g = None
            if f.isDerivedFrom("App::DocumentObjectGroup"):
                g = f
                FreeCADGui.Selection.clearSelection()
                for o in f.OutList:
                    FreeCADGui.Selection.addSelection(o)
            else:
                if f.InList:
                    if f.InList[0].isDerivedFrom("App::DocumentObjectGroup"):
                        g = f.InList[0]
            fast = params.get_param_arch("ConversionFast")
            tol = params.get_param_arch("ConversionTolerance")
            flat = params.get_param_arch("ConversionFlat")
            cut = params.get_param_arch("ConversionCut")
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Mesh to shape"))
            for obj in FreeCADGui.Selection.getSelection():
                newobj = Arch.meshToShape(obj, True, fast, tol, flat, cut)
                if g and newobj:
                    g.addObject(newobj)
            FreeCAD.ActiveDocument.commitTransaction()


class Arch_SelectNonSolidMeshes:
    "the Arch SelectNonSolidMeshes command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_SelectNonManifold.svg",
            "MenuText": QT_TRANSLATE_NOOP(
                "Arch_SelectNonSolidMeshes", "Select Non-Manifold Meshes"
            ),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_SelectNonSolidMeshes",
                "Selects all non-manifold meshes from the document or from the selected groups",
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        msel = []
        if FreeCADGui.Selection.getSelection():
            for o in FreeCADGui.Selection.getSelection():
                if o.isDerivedFrom("App::DocumentObjectGroup"):
                    msel.extend(o.OutList)
        if not msel:
            msel = FreeCAD.ActiveDocument.Objects
        sel = []
        for o in msel:
            if o.isDerivedFrom("Mesh::Feature"):
                if (not o.Mesh.isSolid()) or o.Mesh.hasNonManifolds():
                    sel.append(o)
        if sel:
            FreeCADGui.Selection.clearSelection()
            for o in sel:
                FreeCADGui.Selection.addSelection(o)


class Arch_RemoveShape:
    "the Arch RemoveShape command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_RemoveShape",
            "MenuText": QT_TRANSLATE_NOOP("Arch_RemoveShape", "Remove Shape From BIM"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_RemoveShape", "Removes cubic shapes from BIM components"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and bool(FreeCADGui.Selection.getSelection())

    def Activated(self):
        import Arch

        sel = FreeCADGui.Selection.getSelection()
        Arch.removeShape(sel)


class Arch_CloseHoles:
    "the Arch CloseHoles command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_CloseHoles",
            "MenuText": QT_TRANSLATE_NOOP("Arch_CloseHoles", "Close Holes"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_CloseHoles", "Closes holes in open shapes, turning them into solids"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and bool(FreeCADGui.Selection.getSelection())

    def Activated(self):
        import Arch

        for o in FreeCADGui.Selection.getSelection():
            s = Arch.closeHole(o.Shape)
            if s:
                o.Shape = s


class Arch_Check:
    "the Arch Check command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_Check",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Check", "Check"),
            "ToolTip": QT_TRANSLATE_NOOP("Arch_Check", "Checks the selected objects for problems"),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and bool(FreeCADGui.Selection.getSelection())

    def Activated(self):
        import Arch

        result = Arch.check(FreeCADGui.Selection.getSelection())
        if not result:
            FreeCAD.Console.PrintMessage(str(translate("Arch", "No problems found!")))
        else:
            FreeCADGui.Selection.clearSelection()
            for i in result:
                FreeCAD.Console.PrintWarning(
                    "Object " + i[0].Name + " (" + i[0].Label + ") " + i[1]
                )
                FreeCADGui.Selection.addSelection(i[0])


class Arch_Survey:
    "the Arch Survey command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_Survey",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Survey", "Survey"),
            "ToolTip": QT_TRANSLATE_NOOP("Arch_Survey", "Starts survey"),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        FreeCADGui.addModule("Arch")
        FreeCADGui.doCommandGui("Arch.survey()")


class Arch_ToggleIfcBrepFlag:
    "the Toggle IFC B-rep flag command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_ToggleIfcBrepFlag",
            "MenuText": QT_TRANSLATE_NOOP("Arch_ToggleIfcBrepFlag", "Toggle IFC B-Rep Flag"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_ToggleIfcBrepFlag", "Forces an object to be exported as B-rep or not"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and bool(FreeCADGui.Selection.getSelection())

    def Activated(self):
        import Arch

        for o in FreeCADGui.Selection.getSelection():
            Arch.toggleIfcBrepFlag(o)


class Arch_Component:
    "the Arch Component command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_Component",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Component", "Component"),
            "Accel": "C, M",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_Component", "Creates an undefined architectural component"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Component"))
            FreeCADGui.addModule("Arch")
            FreeCADGui.addModule("Draft")
            FreeCADGui.Control.closeDialog()
            for o in sel:
                FreeCADGui.doCommand(
                    "obj = Arch.makeComponent(FreeCAD.ActiveDocument." + o.Name + ")"
                )
                FreeCADGui.doCommand("Draft.autogroup(obj)")
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()


class Arch_CloneComponent:
    "the Arch Clone Component command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_Component_Clone",
            "MenuText": QT_TRANSLATE_NOOP("Arch_CloneComponent", "Clone Component"),
            "Accel": "C, C",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_CloneComponent", "Clones an object as an undefined architectural component"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        sel = FreeCADGui.Selection.getSelection()
        if sel:
            FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Create Component"))
            FreeCADGui.addModule("Arch")
            FreeCADGui.addModule("Draft")
            FreeCADGui.Control.closeDialog()
            for o in sel:
                FreeCADGui.doCommand(
                    "obj = Arch.cloneComponent(FreeCAD.ActiveDocument." + o.Name + ")"
                )
                FreeCADGui.doCommand("Draft.autogroup(obj)")
            FreeCAD.ActiveDocument.commitTransaction()
            FreeCAD.ActiveDocument.recompute()


class Arch_IfcSpreadsheet:
    "the Arch Schedule command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_Schedule",
            "MenuText": QT_TRANSLATE_NOOP("Arch_IfcSpreadsheet", "New IFC Spreadsheet"),
            "Accel": "I, S",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_IfcSpreadsheet", "Creates a spreadsheet to store IFC properties of an object"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        sel = FreeCADGui.Selection.getSelection()
        FreeCAD.ActiveDocument.openTransaction(
            translate("Arch", "Create IFC properties spreadsheet")
        )
        FreeCADGui.addModule("Arch")
        FreeCADGui.Control.closeDialog()
        if sel:
            for o in sel:
                FreeCADGui.doCommand(
                    "Arch.makeIfcSpreadsheet(FreeCAD.ActiveDocument." + o.Name + ")"
                )
        else:
            FreeCADGui.doCommand("Arch.makeIfcSpreadsheet()")
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


class Arch_ToggleSubs:
    "the ToggleSubs command definition"

    def GetResources(self):
        return {
            "Pixmap": "Arch_ToggleSubs",
            "Accel": "Ctrl+Space",
            "MenuText": QT_TRANSLATE_NOOP("Arch_ToggleSubs", "Toggle Subcomponents"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_ToggleSubs", "Shows or hides the subcomponents of this object"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and bool(FreeCADGui.Selection.getSelection())

    def Activated(self):
        import Draft

        mode = None
        for obj in FreeCADGui.Selection.getSelection():
            if hasattr(obj, "Subtractions"):
                for sub in obj.Subtractions:
                    if not (Draft.getType(sub) in ["Window", "Roof"]):
                        if mode is None:
                            # take the first sub as base
                            mode = sub.ViewObject.isVisible()
                        if mode:
                            sub.ViewObject.hide()
                        else:
                            sub.ViewObject.show()


class Arch_MergeWalls:
    """The command definition for the Arch workbench's gui tool, Arch MergeWalls.

    A tool for merging walls.

    Join two or more walls by using the ArchWall.joinWalls() function.

    Find documentation on the end user usage of Arch Wall here:
    https://wiki.freecad.org/Arch_MergeWalls
    """

    def GetResources(self):
        """Returns a dictionary with the visual aspects of the Arch MergeWalls tool."""

        return {
            "Pixmap": "Arch_MergeWalls",
            "MenuText": QT_TRANSLATE_NOOP("Arch_MergeWalls", "Merge Walls"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Arch_MergeWalls", "Merges the selected walls, if possible"
            ),
        }

    def IsActive(self):
        """Determines whether or not the Arch MergeWalls tool is active.

        Inactive commands are indicated by a greyed-out icon in the menus and
        toolbars.
        """

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v and bool(FreeCADGui.Selection.getSelection())

    def Activated(self):
        """Executed when Arch MergeWalls is called.

        Call ArchWall.joinWalls() on walls selected by the user, with the
        delete option enabled. If the user has selected a single wall, check to
        see if the wall has any Additions that are walls. If so, merges these
        additions to the wall, deleting the additions.
        """

        import Draft
        import ArchWall

        walls = FreeCADGui.Selection.getSelection()
        if len(walls) == 1:
            if Draft.getType(walls[0]) == "Wall":
                ostr = "FreeCAD.ActiveDocument." + walls[0].Name
                ok = False
                for o in walls[0].Additions:
                    if Draft.getType(o) == "Wall":
                        ostr += ",FreeCAD.ActiveDocument." + o.Name
                        ok = True
                if ok:
                    FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Merge Walls"))
                    FreeCADGui.addModule("Arch")
                    FreeCADGui.doCommand("Arch.joinWalls([" + ostr + "],delete=True)")
                    FreeCAD.ActiveDocument.commitTransaction()
                    return
                else:
                    FreeCAD.Console.PrintWarning(
                        translate("Arch", "The selected wall contains no subwalls to merge")
                    )
                    return
            else:
                FreeCAD.Console.PrintWarning(translate("Arch", "Select only wall objects"))
                return
        for w in walls:
            if Draft.getType(w) != "Wall":
                FreeCAD.Console.PrintMessage(translate("Arch", "Select only wall objects"))
                return
        if not ArchWall.areSameWallTypes(walls):
            FreeCAD.Console.PrintMessage(
                translate(
                    "Arch",
                    "Walls with different 'Width', 'Height' and 'Align' properties cannot be merged",
                )
            )
            return
        FreeCAD.ActiveDocument.openTransaction(translate("Arch", "Merge Walls"))
        FreeCADGui.addModule("Arch")
        FreeCADGui.doCommand("Arch.joinWalls(FreeCADGui.Selection.getSelection(),delete=True)")
        FreeCAD.ActiveDocument.commitTransaction()


FreeCADGui.addCommand("Arch_Add", Arch_Add())
FreeCADGui.addCommand("Arch_Remove", Arch_Remove())
FreeCADGui.addCommand("Arch_SplitMesh", Arch_SplitMesh())
FreeCADGui.addCommand("Arch_MeshToShape", Arch_MeshToShape())
FreeCADGui.addCommand("Arch_SelectNonSolidMeshes", Arch_SelectNonSolidMeshes())
FreeCADGui.addCommand("Arch_RemoveShape", Arch_RemoveShape())
FreeCADGui.addCommand("Arch_CloseHoles", Arch_CloseHoles())
FreeCADGui.addCommand("Arch_Check", Arch_Check())
FreeCADGui.addCommand("Arch_Survey", Arch_Survey())
FreeCADGui.addCommand("Arch_ToggleIfcBrepFlag", Arch_ToggleIfcBrepFlag())
FreeCADGui.addCommand("Arch_Component", Arch_Component())
FreeCADGui.addCommand("Arch_CloneComponent", Arch_CloneComponent())
FreeCADGui.addCommand("Arch_IfcSpreadsheet", Arch_IfcSpreadsheet())
FreeCADGui.addCommand("Arch_ToggleSubs", Arch_ToggleSubs())
FreeCADGui.addCommand("Arch_MergeWalls", Arch_MergeWalls())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from lazy_loader.lazy_loader import LazyLoader
from PySide import QtCore, QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import FreeCADGui
import Path
import Path.Base.Gui.Util as PathGuiUtil
import Path.Base.Util as PathUtil
import Path.Tool.Controller as PathToolController
from Path.Tool.toolbit.ui.selector import ToolBitSelector

Part = LazyLoader("Part", globals(), "Part")


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


class ViewProvider:
    def __init__(self, vobj):
        vobj.Proxy = self
        self.vobj = vobj

    def attach(self, vobj):
        mode = 2
        vobj.setEditorMode("LineWidth", mode)
        vobj.setEditorMode("MarkerColor", mode)
        vobj.setEditorMode("NormalColor", mode)
        vobj.setEditorMode("DisplayMode", mode)
        vobj.setEditorMode("BoundingBox", mode)
        vobj.setEditorMode("Selectable", mode)
        vobj.setEditorMode("ShapeAppearance", mode)
        vobj.setEditorMode("Transparency", mode)
        vobj.setEditorMode("Visibility", mode)
        self.vobj = vobj

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def getIcon(self):
        return ":/icons/CAM_ToolController.svg"

    def onChanged(self, vobj, prop):
        mode = 2
        vobj.setEditorMode("LineWidth", mode)
        vobj.setEditorMode("MarkerColor", mode)
        vobj.setEditorMode("NormalColor", mode)
        vobj.setEditorMode("DisplayMode", mode)
        vobj.setEditorMode("BoundingBox", mode)
        vobj.setEditorMode("Selectable", mode)

    def onDelete(self, vobj, args=None):
        PathUtil.clearExpressionEngine(vobj.Object)
        self.vobj.Object.Proxy.onDelete(vobj.Object, args)
        return True

    def updateData(self, vobj, prop):
        # this is executed when a property of the APP OBJECT changes
        pass

    def setEdit(self, vobj=None, mode=0):
        if 0 == mode:
            if vobj is None:
                vobj = self.vobj
            FreeCADGui.Control.closeDialog()
            taskd = TaskPanel(vobj.Object)
            FreeCADGui.Control.showDialog(taskd)
            taskd.setupUi()

            FreeCAD.ActiveDocument.recompute()

            return True
        return False

    def unsetEdit(self, vobj, mode):
        # this is executed when the user cancels or terminates edit mode
        return False

    def setupContextMenu(self, vobj, menu):
        Path.Log.track()
        for action in menu.actions():
            menu.removeAction(action)
        action = QtGui.QAction(translate("CAM", "Edit"), menu)
        action.triggered.connect(self._editInContextMenuTriggered)
        menu.addAction(action)

    def _editInContextMenuTriggered(self, checked):
        self.setEdit()

    def claimChildren(self):
        obj = self.vobj.Object
        if obj and obj.Proxy and obj.Tool:
            return [obj.Tool]
        return []


def Create(name="Default Tool", tool=None, toolNumber=1):
    Path.Log.track(tool, toolNumber)

    obj = PathToolController.Create(name, tool, toolNumber)
    ViewProvider(obj.ViewObject)
    # ToolBits are visible by default, which is typically not what the user wants
    if tool and tool.ViewObject and tool.ViewObject.Visibility:
        tool.ViewObject.Visibility = False
    return obj


class CommandPathToolController(object):
    def GetResources(self):
        return {
            "Pixmap": "CAM_LengthOffset",
            "MenuText": QT_TRANSLATE_NOOP("CAM_ToolController", "Tool Controller"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_ToolController", "Adds a new tool controller to the active job"
            ),
        }

    def selectedJob(self):
        if FreeCAD.ActiveDocument:
            sel = FreeCADGui.Selection.getSelectionEx()
            if sel and sel[0].Object.Name[:3] == "Job":
                return sel[0].Object
        jobs = [o for o in FreeCAD.ActiveDocument.Objects if o.Name[:3] == "Job"]
        if 1 == len(jobs):
            return jobs[0]
        return None

    def IsActive(self):
        return self.selectedJob() is not None

    def Activated(self):
        Path.Log.track()
        job = self.selectedJob()
        if not job:
            return

        # Let the user select a toolbit
        selector = ToolBitSelector()
        if not selector.exec_():
            return
        tool = selector.get_selected_tool()
        if not tool:
            return

        # Find a tool number
        toolNr = None
        for tc in job.Tools.Group:
            if tc.Tool == tool:
                toolNr = tc.ToolNumber
                break
        if not toolNr:
            toolNr = max([tc.ToolNumber for tc in job.Tools.Group]) + 1

        # Create the new tool controller with the tool.
        tc = Create("TC: {}".format(tool.Label), tool, toolNr)
        job.Proxy.addToolController(tc)
        FreeCAD.ActiveDocument.recompute()


class BlockScrollWheel(QtCore.QObject):
    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.Type.Wheel:
            if not obj.hasFocus():
                return True
        return super().eventFilter(obj, event)


class ToolControllerEditor(object):

    def __init__(
        self, obj, asDialog, notifyChanged=None, showCountLabel=False, disableToolNumber=False
    ):
        self.notifyChanged = notifyChanged
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/DlgToolControllerEdit.ui")
        self.controller = FreeCADGui.PySideUic.loadUi(":/panels/ToolControllerEdit.ui")
        self.form.tc_layout.addWidget(self.controller)
        if not asDialog:
            self.form.buttonBox.hide()
        if not showCountLabel:
            self.controller.tcOperationCountLabel.hide()
        self.obj = obj

        comboToPropertyMap = [("spindleDirection", "SpindleDir")]
        enumTups = PathToolController.ToolController.propertyEnumerations(dataType="raw")

        PathGuiUtil.populateCombobox(self.controller, enumTups, comboToPropertyMap)
        self.vertFeed = PathGuiUtil.QuantitySpinBox(self.controller.vertFeed, obj, "VertFeed")
        self.horizFeed = PathGuiUtil.QuantitySpinBox(self.controller.horizFeed, obj, "HorizFeed")
        self.vertRapid = PathGuiUtil.QuantitySpinBox(self.controller.vertRapid, obj, "VertRapid")
        self.horizRapid = PathGuiUtil.QuantitySpinBox(self.controller.horizRapid, obj, "HorizRapid")

        self.blockScrollWheel = BlockScrollWheel()
        self.controller.tcNumber.installEventFilter(self.blockScrollWheel)
        self.controller.spindleDirection.installEventFilter(self.blockScrollWheel)
        self.controller.tcNumber.setReadOnly(disableToolNumber)

        self.editor = None

    def selectInComboBox(self, name, combo):
        """selectInComboBox(name, combo) ...
        helper function to select a specific value in a combo box."""
        try:
            combo.blockSignals(True)
            index = combo.currentIndex()  # Save initial index

            # Search using currentData and return if found
            newindex = combo.findData(name)
            if newindex >= 0:
                combo.setCurrentIndex(newindex)
                return

            # if not found, search using current text
            newindex = combo.findText(name, QtCore.Qt.MatchFixedString)
            if newindex >= 0:
                combo.setCurrentIndex(newindex)
                return

            # not found, return unchanged
            combo.setCurrentIndex(index)
        finally:
            combo.blockSignals(False)

    def updateUi(self):
        tc = self.obj

        blockObjects = [
            self.controller.tcName,
            self.controller.tcNumber,
            self.horizFeed.widget,
            self.horizRapid.widget,
            self.vertFeed.widget,
            self.vertRapid.widget,
            self.controller.spindleSpeed,
            self.controller.spindleDirection,
        ]
        try:
            for obj in blockObjects:
                obj.blockSignals(True)

            self.controller.tcName.setText(tc.Label)
            self.controller.tcNumber.setValue(tc.ToolNumber)
            self.horizFeed.updateWidget()
            self.horizRapid.updateWidget()
            self.vertFeed.updateWidget()
            self.vertRapid.updateWidget()
            self.controller.spindleSpeed.setValue(tc.SpindleSpeed)

            self.selectInComboBox(tc.SpindleDir, self.controller.spindleDirection)

            if self.editor:
                self.editor.updateUI()
        finally:
            for obj in blockObjects:
                obj.blockSignals(False)

    def updateToolController(self):
        tc = self.obj
        try:
            tc.Label = self.controller.tcName.text()
            tc.ToolNumber = self.controller.tcNumber.value()
            self.horizFeed.updateProperty()
            self.vertFeed.updateProperty()
            self.horizRapid.updateProperty()
            self.vertRapid.updateProperty()
            tc.SpindleSpeed = self.controller.spindleSpeed.value()
            tc.SpindleDir = self.controller.spindleDirection.currentData()

            if self.editor:
                self.editor.updateTool()
                tc.Tool = self.editor.tool

        except Exception as e:
            Path.Log.error("Error updating TC: {}".format(e))

    def changed(self):
        self.form.blockSignals(True)
        self.controller.blockSignals(True)
        self.updateToolController()
        self.updateUi()
        self.controller.blockSignals(False)
        self.form.blockSignals(False)

        if self.notifyChanged:
            self.notifyChanged()

    def setupUi(self):
        if self.editor:
            self.editor.setupUI()

        self.controller.tcName.textChanged.connect(self.changed)
        self.controller.tcNumber.editingFinished.connect(self.changed)
        self.vertFeed.widget.textChanged.connect(self.changed)
        self.horizFeed.widget.textChanged.connect(self.changed)
        self.vertRapid.widget.textChanged.connect(self.changed)
        self.horizRapid.widget.textChanged.connect(self.changed)
        self.controller.spindleSpeed.editingFinished.connect(self.changed)
        self.controller.spindleDirection.currentIndexChanged.connect(self.changed)


class TaskPanel:
    def __init__(self, obj):
        self.editor = ToolControllerEditor(obj, False)
        self.form = self.editor.form
        self.updating = False
        self.toolrep = None
        self.obj = obj

    def accept(self):
        self.getFields()

        FreeCADGui.ActiveDocument.resetEdit()
        FreeCADGui.Control.closeDialog()
        if self.toolrep:
            FreeCAD.ActiveDocument.removeObject(self.toolrep.Name)
        FreeCAD.ActiveDocument.recompute()

    def reject(self):
        FreeCADGui.Control.closeDialog()
        if self.toolrep:
            FreeCAD.ActiveDocument.removeObject(self.toolrep.Name)
        FreeCAD.ActiveDocument.recompute()

    def getFields(self):
        self.editor.updateToolController()
        self.obj.Proxy.execute(self.obj)

    def setFields(self):
        self.editor.updateUi()

        if self.toolrep:
            tool = self.obj.Tool
            radius = float(tool.Diameter) / 2
            length = tool.CuttingEdgeHeight
            t = Part.makeCylinder(radius, length)
            self.toolrep.Shape = t

    def edit(self, item, column):
        if not self.updating:
            self.resetObject()

    def resetObject(self, remove=None):
        "transfers the values from the widget to the object"
        FreeCAD.ActiveDocument.recompute()

    def setupUi(self):
        if self.editor.editor:
            t = Part.makeCylinder(1, 1)
            self.toolrep = FreeCAD.ActiveDocument.addObject("Part::Feature", "tool")
            self.toolrep.Shape = t

        self.setFields()
        self.editor.setupUi()


class DlgToolControllerEdit:
    def __init__(self, obj):
        self.editor = ToolControllerEditor(obj, True)
        self.editor.updateUi()
        self.editor.setupUi()
        self.obj = obj

    def exec_(self):
        restoreTC = self.obj.Proxy.templateAttrs(self.obj)

        rc = False
        if not self.editor.form.exec_():
            Path.Log.info("revert")
            self.obj.Proxy.setFromTemplate(self.obj, restoreTC)
            rc = True
        else:
            self.editor.updateToolController()
            self.obj.Proxy.execute(self.obj)
        return rc


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_ToolController", CommandPathToolController())

FreeCAD.Console.PrintLog("Loading PathToolControllerGui… done\n")



from Extension import Extension


class DocumentObjectExtension(Extension):
    """
    Base class for all document object extensions
    Author: Stefan Troeger (stefantroeger@gmx.net)
    Licence: LGPL
    """
    ...



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************


from PySide import QtWidgets
import FreeCADGui
import FreeCAD
from ..models.library import Library


class LibraryPropertyDialog(QtWidgets.QDialog):
    def __init__(self, library: Library, new=False, parent=None):
        super(LibraryPropertyDialog, self).__init__(parent)
        self.library = library

        # Load the UI file into a QWidget
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/LibraryProperties.ui")

        # Create a layout for the dialog and add the loaded form widget
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self.form)
        self.setLayout(layout)

        # Connect signals and set initial values using the loaded form
        self.form.lineEditLibraryName.setText(self.library.label)
        self.update_window_title()

        self.form.buttonBox.accepted.connect(self.save_properties)
        self.form.buttonBox.rejected.connect(self.reject)

        # Connect text changed signal to update window title
        self.form.lineEditLibraryName.textChanged.connect(self.update_window_title)

        # Make the OK button the default so Enter key works
        ok_button = self.form.buttonBox.button(QtWidgets.QDialogButtonBox.Ok)
        cancel_button = self.form.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel)

        if cancel_button:
            cancel_button.setDefault(False)
            cancel_button.setAutoDefault(False)

        if ok_button:
            ok_button.setDefault(True)
            ok_button.setAutoDefault(True)
            ok_button.setFocus()  # Also set focus to the OK button

        # Set minimum width for the dialog
        self.setMinimumWidth(450)

        # Set focus to the text input so user can start typing immediately
        self.form.lineEditLibraryName.setFocus()
        self.form.lineEditLibraryName.selectAll()  # Select all text for easy replacement

    def update_window_title(self):
        # Update title based on current text in the line edit
        current_name = self.form.lineEditLibraryName.text()
        title = FreeCAD.Qt.translate(
            "LibraryPropertyDialog", f"Library Properties - {current_name or self.library.label}"
        )
        self.setWindowTitle(title)

    def save_properties(self):
        new_name = self.form.lineEditLibraryName.text()
        if new_name != self.library.label:
            self.library._label = new_name
        # Additional logic to save other properties if added later
        self.accept()



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

import functools
import os
import subprocess
import tempfile
from time import sleep
import unittest

from addonmanager_dependency_installer import DependencyInstaller


class CompleteProcessMock(subprocess.CompletedProcess):
    def __init__(self):
        super().__init__(["fake_arg"], 0)
        self.stdout = "Mock subprocess call stdout result"


class SubprocessMock:
    def __init__(self):
        self.arg_log = []
        self.called = False
        self.call_count = 0
        self.delay = 0
        self.succeed = True

    def subprocess_interceptor(self, args):
        self.arg_log.append(args)
        self.called = True
        self.call_count += 1
        sleep(self.delay)
        if self.succeed:
            return CompleteProcessMock()
        raise subprocess.CalledProcessError(1, " ".join(args), "Unit test mock output")


class FakeFunction:
    def __init__(self):
        self.called = False
        self.call_count = 0
        self.return_value = None
        self.arg_log = []

    def func_call(self, *args):
        self.arg_log.append(args)
        self.called = True
        self.call_count += 1
        return self.return_value


class TestDependencyInstaller(unittest.TestCase):
    """Test the dependency installation class"""

    def setUp(self):
        self.subprocess_mock = SubprocessMock()
        self.test_object = DependencyInstaller([], ["required_py_package"], ["optional_py_package"])
        self.test_object._subprocess_wrapper = self.subprocess_mock.subprocess_interceptor
        self.signals_caught = []
        self.test_object.failure.connect(functools.partial(self.catch_signal, "failure"))
        self.test_object.finished.connect(functools.partial(self.catch_signal, "finished"))
        self.test_object.no_pip.connect(functools.partial(self.catch_signal, "no_pip"))
        self.test_object.no_python_exe.connect(
            functools.partial(self.catch_signal, "no_python_exe")
        )

    def tearDown(self):
        pass

    def catch_signal(self, signal_name, *_):
        self.signals_caught.append(signal_name)

    def test_run_no_pip(self):
        self.test_object._verify_pip = lambda: False
        self.test_object.run()
        self.assertIn("finished", self.signals_caught)

    def test_run_with_pip(self):
        ff = FakeFunction()
        self.test_object._verify_pip = lambda: True
        self.test_object._install_python_packages = ff.func_call
        self.test_object.run()
        self.assertIn("finished", self.signals_caught)
        self.assertTrue(ff.called)

    def test_run_with_no_packages(self):
        ff = FakeFunction()
        self.test_object._verify_pip = lambda: True
        self.test_object._install_python_packages = ff.func_call
        self.test_object.python_requires = []
        self.test_object.python_optional = []
        self.test_object.run()
        self.assertIn("finished", self.signals_caught)
        self.assertFalse(ff.called)

    def test_install_python_packages_new_location(self):
        ff_required = FakeFunction()
        ff_optional = FakeFunction()
        self.test_object._install_required = ff_required.func_call
        self.test_object._install_optional = ff_optional.func_call
        with tempfile.TemporaryDirectory() as td:
            self.test_object.location = os.path.join(td, "UnitTestLocation")
            self.test_object._install_python_packages()
            self.assertTrue(ff_required.called)
            self.assertTrue(ff_optional.called)
            self.assertTrue(os.path.exists(self.test_object.location))

    def test_install_python_packages_existing_location(self):
        ff_required = FakeFunction()
        ff_optional = FakeFunction()
        self.test_object._install_required = ff_required.func_call
        self.test_object._install_optional = ff_optional.func_call
        with tempfile.TemporaryDirectory() as td:
            self.test_object.location = td
            self.test_object._install_python_packages()
            self.assertTrue(ff_required.called)
            self.assertTrue(ff_optional.called)

    def test_verify_pip_no_pip(self):
        sm = SubprocessMock()
        sm.succeed = False
        self.test_object._subprocess_wrapper = sm.subprocess_interceptor
        self.test_object._get_python = lambda: "fake_python"
        result = self.test_object._verify_pip()
        self.assertFalse(result)
        self.assertIn("no_pip", self.signals_caught)

    def test_verify_pip_with_pip(self):
        sm = SubprocessMock()
        sm.succeed = True
        self.test_object._subprocess_wrapper = sm.subprocess_interceptor
        self.test_object._get_python = lambda: "fake_python"
        result = self.test_object._verify_pip()
        self.assertTrue(result)
        self.assertNotIn("no_pip", self.signals_caught)

    def test_install_required_loops(self):
        sm = SubprocessMock()
        sm.succeed = True
        self.test_object._subprocess_wrapper = sm.subprocess_interceptor
        self.test_object._get_python = lambda: "fake_python"
        self.test_object.python_requires = ["test1", "test2", "test3"]
        self.test_object._install_required("vendor_path")
        self.assertEqual(sm.call_count, 3)

    def test_install_required_failure(self):
        sm = SubprocessMock()
        sm.succeed = False
        self.test_object._subprocess_wrapper = sm.subprocess_interceptor
        self.test_object._get_python = lambda: "fake_python"
        self.test_object.python_requires = ["test1", "test2", "test3"]
        self.test_object._install_required("vendor_path")
        self.assertEqual(sm.call_count, 1)
        self.assertIn("failure", self.signals_caught)

    def test_install_optional_loops(self):
        sm = SubprocessMock()
        sm.succeed = True
        self.test_object._subprocess_wrapper = sm.subprocess_interceptor
        self.test_object._get_python = lambda: "fake_python"
        self.test_object.python_optional = ["test1", "test2", "test3"]
        self.test_object._install_optional("vendor_path")
        self.assertEqual(sm.call_count, 3)

    def test_install_optional_failure(self):
        sm = SubprocessMock()
        sm.succeed = False
        self.test_object._subprocess_wrapper = sm.subprocess_interceptor
        self.test_object._get_python = lambda: "fake_python"
        self.test_object.python_optional = ["test1", "test2", "test3"]
        self.test_object._install_optional("vendor_path")
        self.assertEqual(sm.call_count, 3)

    def test_run_pip(self):
        pass



# SPDX-License-Identifier: LGPL-2.1-or-later

from Base.Metadata import export
from Base.BaseClass import BaseClass
from typing import Any, Final, Tuple

@export(
    Include="Gui/Selection/SelectionObject.h",
    Delete=True,
)
class SelectionObject(BaseClass):
    """
    This class represents selections made by the user. It holds information about the object, document and sub-element of the selection.

    Author: Juergen Riegel (FreeCAD@juergen-riegel.net)
    Licence: LGPL
    """

    def remove(self) -> None:
        """
        Remove this selection item from the selection.
        remove() -> None
        --
        This object becomes invalid.
        """
        ...

    def isObjectTypeOf(self, type: Any) -> bool:
        """
        Test for a certain father class.
        isObjectTypeOf(type) -> Bool
        """
        ...
    ObjectName: Final[str] = ""
    """Name of the selected object"""

    SubElementNames: Final[Tuple[str, ...]] = ()
    """Name of the selected sub-element if any"""

    FullName: Final[str] = ""
    """Name of the selected object"""

    TypeName: Final[str] = ""
    """Type name of the selected object"""

    DocumentName: Final[str] = ""
    """Name of the document of the selected object"""

    Document: Final[Any] = ...
    """Document of the selected object"""

    Object: Final[Any] = ...
    """Selected object"""

    SubObjects: Final[Tuple[Any, ...]] = ()
    """Selected sub-element, if any"""

    PickedPoints: Final[Tuple[Any, ...]] = ()
    """Picked points for selection"""

    HasSubObjects: Final[bool] = False
    """Selected sub-element, if any"""
